#include "stdafx.h"
#include "resource.h"
#include "JpgView.h"

#include <commdlg.h>
#include <math.h>

#define MAX_LOADSTRING		256
#define BUF_SIZE			0x10000

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK About(HWND, UINT, WPARAM, LPARAM);
void DrawBitmap(HDC);
void UpdateScrollers(bool);
void ScrollImage(int, int, bool);
void OpenFile();
void SaveFile();
void ReadBytes(BYTE *, int);
void OpenJPEGFile(char *);
void SaveBMPFile(char *);
inline void Swap(WORD&);
void ReadApp0(BYTE *, WORD);
void ReadDQT(BYTE *, WORD);
void ReadDHT(BYTE *, WORD);
void ReadSof0(BYTE *, WORD);
void ReadSos(BYTE *, WORD);
void DecodeJPEG();
void BuildDQTMatrix();
void BuildHuffmanTree(int);
HTree *AddNode(HTree *&, int, BYTE);
void DecodeMCU(DiscrInfo *, int, BYTE *);
void ReadDU(int [][8], int);
BYTE GetHuffmanCode(int);
inline void Invert(int&, int);
int ReadBits(int);
inline bool Read1Bit();
void ReadByte();
void ProcessDU(int [][8], int);
void TransposeMatrix(int [][8]);
void DeQuantMatrix(int [][8], int [][8]);
void IDQT(int [][8]);

HINSTANCE hInst;
HWND hWnd;
TCHAR szTitle[MAX_LOADSTRING];
TCHAR szWindowClass[MAX_LOADSTRING];
char FileName[MAX_PATH] = "";

HANDLE hFile;
HBITMAP hBm = 0;

int CurBit;
BYTE CurByte;
DWORD MCUInterval, MCUNo;

APP0Struct App0;
SOF0Struct Sof0;
SOSStruct Sos;

HuffmanTable HTable[4];
int QTable[4][8][8];
double C[8][8];
HTree *Ht[4];
int DC[3];
int Hi[4];

BYTE *FileBuf;
DWORD FilePos;

int ScrollX = 0, ScrollY = 0;

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	WNDCLASS wc;
	MSG msg;

	BuildDQTMatrix();

	LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
	LoadString(hInstance, IDC_JPGVIEW, szWindowClass, MAX_LOADSTRING);

	ZeroMemory(&wc, sizeof(WNDCLASS));
	wc.lpfnWndProc = (WNDPROC)WndProc;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wc.lpszMenuName = (LPCSTR)IDC_JPGVIEW;
	wc.lpszClassName = szWindowClass;

	RegisterClass(&wc);

	hInst = hInstance;

	hWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW | WS_HSCROLL | WS_VSCROLL,
		CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

	if (!hWnd)
	{
		return FALSE;
	}

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	while (GetMessage(&msg, NULL, 0, 0)) 
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	HDC hdc;

	switch (message) 
	{
		case WM_COMMAND:
			wmId = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			switch (wmId)
			{
				case IDM_FILEOPEN:
					OpenFile();
					break;

				case IDM_FILESAVEAS:
					SaveFile();
					break;

				case IDM_EXIT:
				   DestroyWindow(hWnd);
				   break;

				case IDM_ABOUT:
				   DialogBox(hInst, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
				   break;
				
				default:
				   return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;

		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			DrawBitmap(hdc);
			EndPaint(hWnd, &ps);
			break;

		case WM_SIZE:
			UpdateScrollers(false);
			break;

		case WM_HSCROLL:
			ScrollImage(LOWORD(wParam), HIWORD(wParam), true);
			break;

		case WM_VSCROLL:
			ScrollImage(LOWORD(wParam), HIWORD(wParam), false);
			break;

		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
	}
   
	return 0;
}

LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:	
			HFONT hFont;
			LOGFONT logfont;
			logfont.lfHeight = -14;
			logfont.lfWidth = 0;
			logfont.lfEscapement = 0;
			logfont.lfOrientation = 0;
			logfont.lfWeight = FW_NORMAL;
			logfont.lfItalic = FALSE;
			logfont.lfUnderline = FALSE;
			logfont.lfStrikeOut = FALSE;
			logfont.lfCharSet = ANSI_CHARSET;
			logfont.lfOutPrecision = OUT_DEFAULT_PRECIS;
			logfont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
			logfont.lfQuality = DEFAULT_QUALITY;
			logfont.lfPitchAndFamily = DEFAULT_PITCH;
			strcpy(logfont.lfFaceName,"MS Sans Serif");
			hFont = CreateFontIndirect(&logfont);
			SendMessage(GetDlgItem(hDlg, IDOK), WM_SETFONT, (DWORD)hFont, 1);
			SendMessage(GetDlgItem(hDlg, IDC_TEXT1), WM_SETFONT, (DWORD)hFont, 1);
			SendMessage(GetDlgItem(hDlg, IDC_TEXT2), WM_SETFONT, (DWORD)hFont, 1);
			return TRUE;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
			{
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
			}
			break;
	}

    return FALSE;
}

void DrawBitmap(HDC hdc)
{
	if (!hBm) return;
	RECT r;
	HDC MemDC = CreateCompatibleDC(hdc);
	SelectObject(MemDC, hBm);
	GetClientRect(hWnd, &r);
	BitBlt(hdc, 0, 0, r.right - r.left, r.bottom - r.top, MemDC, ScrollX, ScrollY, SRCCOPY);
	DeleteDC(MemDC);
}

void UpdateScrollers(bool ZeroPos)
{
	SCROLLINFO si;
	RECT r;

	int W = hBm ? Sof0.Width : 0;
	int H = hBm ? Sof0.Height : 0;

	GetClientRect(hWnd, &r);

	si.fMask = SIF_PAGE | SIF_RANGE;
	si.cbSize = sizeof(SCROLLINFO);
	si.nMax = W - 1;
	si.nMin = 0;
	si.nPage = r.right - r.left;
	if (ZeroPos)
	{
		si.fMask |= SIF_POS;
		si.nPos = 0;
	}

	SetScrollInfo(hWnd, SB_HORZ, &si, TRUE);

	si.nMax = H - 1;
	si.nPage = r.bottom - r.top;
	if (ZeroPos)
	{
		si.fMask |= SIF_POS;
		si.nPos = 0;
	}

	SetScrollInfo(hWnd, SB_VERT, &si, TRUE);

	W = GetScrollPos(hWnd, SB_HORZ);
	H = GetScrollPos(hWnd, SB_VERT);
	if (ScrollX != W || ScrollY != H) 
	{
		ScrollX = W;
		ScrollY = H;
		InvalidateRect(hWnd, NULL, FALSE);
	}
}

void ScrollImage(int Code, int nPos, bool Horiz)
{
	RECT r;
	int k;
	
	GetClientRect(hWnd, &r);
	
	if (Horiz)
	{
		k = r.right - r.left;
		switch (Code)
		{
			case SB_LINELEFT: ScrollX -= 8; break;
			case SB_LINERIGHT: ScrollX += 8; break;
			case SB_THUMBPOSITION:
			case SB_THUMBTRACK: ScrollX = nPos; break;
			case SB_PAGERIGHT: ScrollX += k; break;
			case SB_PAGELEFT: ScrollX -= k;
		}
		SetScrollPos(hWnd, SB_HORZ, ScrollX, TRUE);
		ScrollX = GetScrollPos(hWnd, SB_HORZ);
	}
	else
	{
		k = r.bottom - r.top;
		switch (Code)
		{
			case SB_LINELEFT: ScrollY -= 8; break;
			case SB_LINERIGHT: ScrollY += 8; break;
			case SB_THUMBPOSITION:
			case SB_THUMBTRACK: ScrollY = nPos; break;
			case SB_PAGERIGHT: ScrollY += k; break;
			case SB_PAGELEFT: ScrollY -= k;
		}
		SetScrollPos(hWnd, SB_VERT, ScrollY, TRUE);
		ScrollY = GetScrollPos(hWnd, SB_VERT);
	}

	InvalidateRect(hWnd, NULL, FALSE);
}

void OpenFile()
{
	OPENFILENAME ofn;

	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd;
	ofn.hInstance = hInst;
	ofn.lpstrFilter = "JPEG (*.jpg)\0*.JPG\0  (*.*)\0*.*\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = FileName;
	ofn.nMaxFile = MAX_PATH;
	ofn.lpstrTitle = "";
	ofn.Flags = 
		OFN_ENABLESIZING | 
		OFN_EXPLORER | 
		OFN_FILEMUSTEXIST | 
		OFN_HIDEREADONLY | 
		OFN_PATHMUSTEXIST;
	ofn.lpstrDefExt = "*.jpg";
	
	if (GetOpenFileName(&ofn))
	{
		EnableMenuItem(GetMenu(hWnd), IDM_FILESAVEAS, MF_GRAYED);
		OpenJPEGFile(FileName);
		if (hBm) 		
			EnableMenuItem(GetMenu(hWnd), IDM_FILESAVEAS, MF_ENABLED);
		ScrollX = ScrollY = 0;
		UpdateScrollers(true);
		InvalidateRect(hWnd, NULL, TRUE);
	}
}

void SaveFile()
{
	OPENFILENAME ofn;

	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd;
	ofn.hInstance = hInst;
	ofn.lpstrFilter = " Windows Bitmap (*.bmp)\0*.BMP\0  (*.*)\0*.*\0";
	ofn.lpstrFile = FileName;
	ofn.nMaxFile = MAX_PATH;
	ofn.lpstrTitle = " ";
	ofn.Flags = 
		OFN_ENABLESIZING | 
		OFN_EXPLORER | 
		OFN_OVERWRITEPROMPT |
		OFN_HIDEREADONLY | 
		OFN_PATHMUSTEXIST;
	ofn.lpstrDefExt = "*.bmp";
	
	if (GetSaveFileName(&ofn))
		SaveBMPFile(FileName);
}

void ReadBytes(BYTE *buf, int n)
{
	if (FilePos + n >= BUF_SIZE - 0x16)
	{
		DWORD Read;
		SetFilePointer(hFile, FilePos - BUF_SIZE, NULL, FILE_CURRENT);
		ReadFile(hFile, FileBuf, BUF_SIZE, &Read, NULL);
		if (!Read) throw "   !";
		FilePos = 0;
	}
	memcpy(buf, FileBuf + FilePos, n);
	FilePos += n;
}

void OpenJPEGFile(char *FileName)
{
	WORD *buf;
	WORD Seg, Size;
	DWORD Read;
	bool APP0Init = false;
	bool DQTInit = false;
	bool DHTInit = false;
	bool SOF0Init = false;
	bool SOSInit = false;

	if (hBm) 
	{
		DeleteObject(hBm);
		hBm = 0;
	}

	hFile = CreateFile(FileName, GENERIC_READ, 
		FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, 0);

	if (hFile == INVALID_HANDLE_VALUE)
	{
		MessageBox(hWnd, "  !", "", MB_OK | MB_ICONSTOP);
		return;
	}

	FileBuf = new BYTE[BUF_SIZE];
	FilePos = 0;
	MCUInterval = 0;
	buf = new WORD[BUF_SIZE >> 2];
	ReadFile(hFile, FileBuf, BUF_SIZE, &Read, NULL);

	try
	{
		ReadBytes((BYTE *)buf, 2);
		if (buf[0] != SOI) 
			throw "    JPEG !";
		do
		{
			ReadBytes((BYTE *)&Seg, 2);
			if ((Seg & 0xFF) != 0xFF)
				throw "   JPEG!";
			if ((Seg >> 4) >= 0xC1 && (Seg >> 4) <= 0xCA)
				throw "  !";
			ReadBytes((BYTE *)&Size, 2);
			Swap(Size);
			Size -= 2;
			ReadBytes((BYTE *)buf, Size);
			switch (Seg)
			{
				case APP0: 
				case APP1: 
				case APP2: 
				case APP3: 
				case APP4: 
				case APP5: 
				case APP6: 
				case APP7: 
				case APP8: 
				case APP9: 
				case APP10: 
				case APP11: 
				case APP12: 
				case APP13: 
				case APP14: 
					if (APP0Init) break;
					ReadApp0((BYTE *)buf, Size);
					APP0Init = true;
					break;

				case DQT: 
					if (DQTInit) break;
					ReadDQT((BYTE *)buf, Size); 
					DQTInit = true;
					break;

				case DHT: 
					if (DHTInit) break;
					ReadDHT((BYTE *)buf, Size); 
					DHTInit = true;
					break;

				case SOF0: 
					if (SOF0Init) break;
					ReadSof0((BYTE *)buf, Size);
					SOF0Init = true;
					break;

				case DRI:
					Swap(buf[0]);
					MCUInterval = buf[0];
					break;

				case SOS: 
					if (SOSInit) break;
					ReadSos((BYTE *)buf, Size);
					SOSInit = true;
			}
		} while (Seg != SOS);

		if (!APP0Init || !DQTInit || !DHTInit || !SOF0Init || !SOSInit)
			throw "      !";

		DecodeJPEG();
	}
	catch (char *err)
	{
		MessageBox(hWnd, err, "", MB_OK | MB_ICONSTOP);
	}

	delete FileBuf;
	delete buf;

	CloseHandle(hFile);
}

void SaveBMPFile(char *FileName)
{
	BITMAPFILEHEADER bfh;
	BITMAPINFOHEADER bih;
	DWORD Written, ImgSize;
	BYTE *buf;
	HDC hDC;
	int W;

	if (!hBm) return;

	HANDLE hFile = CreateFile(FileName, GENERIC_WRITE, 
		0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);

	if (hFile == INVALID_HANDLE_VALUE)
	{
		MessageBox(hWnd, "  !", "", MB_OK | MB_ICONSTOP);
		return;
	}

	W = Sof0.Width;
	if (W & 0x03) W += 4 - W & 0x03;
	ImgSize = W * Sof0.Height * 3;

	bfh.bfType = 0x4D42;
	bfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	bfh.bfSize = bfh.bfOffBits + ImgSize;
	bfh.bfReserved1 = 0;
	bfh.bfReserved2 = 0;

	ZeroMemory(&bih, sizeof(BITMAPINFOHEADER));
	bih.biSize = sizeof(BITMAPINFOHEADER);
	bih.biWidth = Sof0.Width;
	bih.biHeight = Sof0.Height;
	bih.biBitCount = 24;
	bih.biPlanes = 1;
	bih.biCompression = BI_RGB;
	
	buf = new BYTE[ImgSize];
	hDC = GetDC(hWnd);
	GetDIBits(hDC, hBm, 0, Sof0.Height, buf, (BITMAPINFO *)&bih, DIB_RGB_COLORS);
	ReleaseDC(hWnd, hDC);

	WriteFile(hFile, &bfh, sizeof(BITMAPFILEHEADER), &Written, NULL);
	WriteFile(hFile, &bih, sizeof(BITMAPINFOHEADER), &Written, NULL);
	WriteFile(hFile, buf, ImgSize, &Written, NULL);

	delete buf;
	CloseHandle(hFile);
}

inline void Swap(WORD& val)
{
	val = ((val & 0xFF) << 8) + (val >> 8);
}

void ReadApp0(BYTE *buf, WORD Size)
{
	memcpy(&App0, buf, sizeof(APP0Struct));
	if (lstrcmp(App0.JFIF, "JFIF") && lstrcmp(App0.JFIF, "Exif"))
		throw "  JFIF!";
	if (!lstrcmp(App0.JFIF, "JFIF"))
	{
		if (App0.VerHi != 1)
			throw "    JFIF!";
		if (App0.VerLo > 2)
			throw "   JFIF!";
	}
}

void ReadDQT(BYTE *buf, WORD Size)
{
	BYTE *p = buf;
	BYTE Info;
	int i, j, k, Is16Bit;

	do
	{
		Info = *p++;
		k = Info & 0x0F;
		if (k > 3)
			throw "   !";
		Is16Bit = Info & 0xF0;
		for (i = 0; i < 8; i++)
			for (j = 0; j < 8; j++)
				if (Is16Bit)
				{
					QTable[k][i][j] = *((WORD *)p);
					p += 2;
				}
				else
					QTable[k][i][j] = *p++;
	} while (p - buf < Size);
}

void ReadDHT(BYTE *buf, WORD Size)
{
	BYTE *p = buf;
	BYTE Info;
	int i, k;

	do
	{
		Info = *p++;
		if ((Info & 0x0F) > 3)
			throw "   !";
		if ((Info >> 4) > 1)
			throw "   !";
		k = (Info & 0x0F) | (Info >> 3);
		memcpy(HTable[k].Count, p, 16);
		p += 16;
		HTable[k].TotalCount = 0;
		for (i = 0; i < 16; i++)
			HTable[k].TotalCount += HTable[k].Count[i];
		memcpy(HTable[k].Bytes, p, HTable[k].TotalCount);
		p += HTable[k].TotalCount;
	} while (p - buf < Size);
}

void ReadSof0(BYTE *buf, WORD Size)
{
	memcpy((char *)(&Sof0) + 1, buf, min(Size, sizeof(SOF0Struct))); 
	if (Sof0.Precis != 8)
		throw "  !";
	Swap(Sof0.Height);
	Swap(Sof0.Width);
	if (!Sof0.Height || !Sof0.Width)
		throw "  !";
	if (Sof0.NumComp != 1 && Sof0.NumComp != 3)
		throw "   !";
	for (int i = 0; i < Sof0.NumComp; i++)
	{
		if (!Sof0.Comps[i].Id || Sof0.Comps[i].Id > 3)
			throw "   !";
		Sof0.Comps[i].QTable = (Sof0.Comps[i].QTable & 1) | (Sof0.Comps[i].QTable >> 3);
		if (Sof0.Comps[i].QTable > 3)
			throw "   !";
	}
}

void ReadSos(BYTE *buf, WORD Size)
{
	memcpy(&Sos, buf, min(Size, sizeof(SOSStruct)));
	if (Sos.NumComp != 1 && Sos.NumComp != 3)
		throw "     !";
	for (int i = 0; i < Sos.NumComp; i++)
	{
		if (!Sos.Comps[i].Id || Sos.Comps[i].Id > 3)
			throw "   !";
		Sos.Comps[i].HTable = (Sos.Comps[i].HTable & 1) | (Sos.Comps[i].HTable >> 3);
		if (Sos.Comps[i].HTable > 3)
			throw "   !";
	}
}

void DecodeJPEG()
{
	BYTE *ImgBuf, *p;
	DiscrInfo di;
	int ScanWidth, W, H;
	int i, j, k, m;

	ZeroMemory(DC, sizeof(DC));
	ZeroMemory(&di, sizeof(DiscrInfo));

	MCUNo = 0;
	CurBit = -1;

	for (i = 0; i < 4; i++)
		BuildHuffmanTree(i);

	for (i = 0; i < Sof0.NumComp; i++)
	{
		j = Sof0.Comps[i].DFactor >> 4;
		if (j > di.MaxHDiscr) di.MaxHDiscr = j;
		k = Sof0.Comps[i].DFactor & 0x0F;
		if (k > di.MaxVDiscr) di.MaxVDiscr = k;
		switch (Sof0.Comps[i].Id)
		{
			case 1: di.YBlocks = j * k; break;
			case 2: di.CbBlocks = j * k; break;
			case 3: di.CrBlocks = j * k;
		}
		Hi[i] = Sos.Comps[i].HTable & 1;
	}

	if (!di.MaxHDiscr || di.MaxHDiscr > 4 ||
		!di.MaxVDiscr || di.MaxVDiscr > 4)
		throw "  !";

	di.BlockW = di.MaxHDiscr << 3;
	di.BlockH = di.MaxVDiscr << 3;
	di.BlockSize = di.BlockW * di.BlockH * 3;

	W = Sof0.Width;
	k = W % di.BlockW;
	if (k) W += di.BlockW - k;
	H = Sof0.Height;
	k = H % di.BlockH;
	if (k) H += di.BlockH - k;

	ScanWidth = W * 3;
	p = ImgBuf = new BYTE[ScanWidth * H];

	k = di.BlockW * 3;
	m = (di.BlockH - 1) * ScanWidth;
	for (i = 0; i < Sof0.Height; i += di.BlockH, p = ImgBuf + i * ScanWidth)
		for (j = 0; j < Sof0.Width; j += di.BlockW, p += k)
			DecodeMCU(&di, ScanWidth, p);

	BITMAPINFOHEADER bih;

	ZeroMemory(&bih, sizeof(BITMAPINFOHEADER));
	bih.biSize = sizeof(BITMAPINFOHEADER);
	bih.biWidth = W;
	bih.biHeight = -Sof0.Height;
	bih.biBitCount = 24;
	bih.biPlanes = 1;
	bih.biCompression = BI_RGB;

	HDC hDC = GetDC(hWnd);
	hBm = CreateCompatibleBitmap(hDC, Sof0.Width, Sof0.Height);
	SetDIBits(hDC, hBm, 0, Sof0.Height, ImgBuf, (BITMAPINFO *)&bih, DIB_RGB_COLORS);
	ReleaseDC(hWnd, hDC);

	for (i = 0; i < 4; i++)
		delete Ht[i];

	delete ImgBuf;
}

void BuildDQTMatrix()
{
	int i = 0, j;

	for (; i < 8; i++)
		for (j = 0; j < 8; j++)
			if (!i)
				C[i][j] = 1 / sqrt(8.0);
			else
				C[i][j] = 0.5 * cos((2 * j + 1) * i * 0.19634954);
}

void BuildHuffmanTree(int k)
{
	int i, j, n = 0;
	HTree *q;

	Ht[k] = NULL;
	for (i = 0; i < 16; i++)
		for (j = 0; j < HTable[k].Count[i]; j++)
		{
			q = AddNode(Ht[k], i + 1, HTable[k].Bytes[n++]);
			if (!q)
			{
				delete Ht[k];
				throw "  !";
			}
		}
}

HTree *AddNode(HTree *&t, int l, BYTE b)
{
	if (!t)
	{
		t = new HTree();
	}
	else if (t->IsLeaf) return NULL;
	if (!l)
	{
		t->IsLeaf = true;
		t->Byte = b;
		return t;
	}
	HTree *node = AddNode(t->lnode, l - 1, b);
	if (node) return node;
	return AddNode(t->rnode, l - 1, b);
}

void DecodeMCU(DiscrInfo *di, int ScanWidth, BYTE *buf)
{
	int Y[4][8][8], Cb[4][8][8], Cr[4][8][8];
	int *Yp, *Cbp, *Crp;
	BYTE t[192], *p, *r, *q;
	int R, G, B;
	int i, j, k, m;
	int xy, yy, xcb, ycb, xcr, ycr;

	for (i = 0; i < di->YBlocks; i++)
	{
		ReadDU(Y[i], 0);
		ProcessDU(Y[i], 0);
	}
	for (i = 0; i < di->CbBlocks; i++)
	{
		ReadDU(Cb[i], 1);
		ProcessDU(Cb[i], 1);
	}
	for (i = 0; i < di->CrBlocks; i++)
	{
		ReadDU(Cr[i], 2);
		ProcessDU(Cr[i], 2);
	}

	if (Sof0.NumComp == 3)
	{
		r = buf;
		m = (ScanWidth << 3) - di->BlockW * 3;
		for (i = 0; i < di->MaxVDiscr; i++, r += m)
		{
			yy = ycb = ycr = i;
			yy %= Sof0.Comps[0].DFactor & 0x0F;
			ycb %= Sof0.Comps[1].DFactor & 0x0F;
			ycr %= Sof0.Comps[2].DFactor & 0x0F;
			for (j = 0; j < di->MaxHDiscr; j++, r += 24)
			{
				xy = xcb = xcr = j;
				xy %= Sof0.Comps[0].DFactor >> 4;
				xcb %= Sof0.Comps[1].DFactor >> 4;
				xcr %= Sof0.Comps[2].DFactor >> 4;
				p = t;
				Yp = &Y[(yy << 1) + xy][0][0];
				Cbp = &Cb[(ycb << 1) + xcb][0][0];
				Crp = &Cr[(ycr << 1) + xcr][0][0];
				for (k = 0; k < 64; k++)
				{
					R = int(*Yp + 128 + 1.402 * *Crp);
					G = int(*Yp + 128 - 0.34414 * *Cbp - 0.71414 * *Crp);
					B = int(*Yp + 128 + 1.772 * *Cbp);
					if (R < 0) R = 0;
					if (R > 255) R = 255;
					if (G < 0) G = 0;
					if (G > 255) G = 255;
					if (B < 0) B = 0;
					if (B > 255) B = 255;
					*p++ = B;
					*p++ = G;
					*p++ = R;
					Yp++;
					Cbp++;
					Crp++;
				}
				p = t;
				q = r;
				for (k = 0; k < 8; k++, q += ScanWidth, p += 24)
					memcpy(q, p, 24);
			}
		}
	}
	else
	{
		p = t;
		Yp = &Y[0][0][0];
		for (i = 0; i < 64; i++)
		{
			*Yp += 128;
			if (*Yp < 0) *Yp = 0;
			if (*Yp > 255) *Yp = 255;
			*p++ = *Yp;
			*p++ = *Yp;
			*p++ = *Yp++;
		}
		p = t;
		for (i = 0; i < 8; i++, p += 24, buf += ScanWidth)
			memcpy(buf, p, 24);
	}

	MCUNo++;
	if (MCUNo == MCUInterval)
	{
		if (FileBuf[FilePos] == 0xFF &&
			FileBuf[FilePos + 1] >= 0xD0 && 
			FileBuf[FilePos + 1] <= 0xD7)
		{
			ZeroMemory(DC, sizeof(DC));
			CurBit = 7;
			FilePos += 2;
			ReadBytes(&CurByte, 1);
		}
		MCUNo = 0;
	}
}

void ReadDU(int M[][8], int type)
{
	int t[128], *p;
	int k = 0, n;
	int h = Hi[type];

	n = GetHuffmanCode(h);
	if (n) 
	{
		k = ReadBits(n);
		Invert(k, n);
	}
	DC[type] += k;
	t[0] = DC[type];
	p = t + 1;

	do
	{
		k = GetHuffmanCode(h + 2);
		if (!k)
		{
			memset(p, 0, (64 - (p - t)) << 2);
			break;
		}
		else
		{
			n = k >> 4;
			if (n) 
			{
				memset(p, 0, n << 2);
				p += n;
			}
			n = k & 0x0F;
			if (n) 
			{
				k = ReadBits(n);
				Invert(k, n);
				*p++ = k;
			}
			else *p++ = 0;
		}
	} while (p - t < 64);

	memcpy(M, t, 256);
}

inline void Invert(int& k, int n)
{
	if (k & (1 << (n - 1))) return;
	k = ~k & ((1 << n) - 1);
	k = -k;
}

BYTE GetHuffmanCode(int k)
{
	HTree *p = Ht[k];

	while(!p->IsLeaf)
	{
		if (Read1Bit()) p = p->rnode;
		else p = p->lnode;
		if (!p) throw "  !";
	}
	
	return p->Byte;
}

int ReadBits(int n)
{
	int b = 0;
	
	for (; n; n--)
	{
		b <<= 1;
		if (CurBit < 0)
		{
			ReadByte();
			CurBit = 7;
		}
		b |= (CurByte >> CurBit--) & 1;
	}

	return b;
}

inline bool Read1Bit()
{
	if (CurBit < 0)
	{
		ReadByte();
		CurBit = 7;
	}
	return (CurByte >> CurBit--) & 1;
}

void ReadByte()
{
	ReadBytes(&CurByte, 1);
	if (CurByte == 0xFF)
	{
		ReadBytes(&CurByte, 1);
		if (CurByte == 0) CurByte = 0xFF;
		else if (CurByte == 0xD9) 
			throw "  !";
	}
}

void ProcessDU(int M[][8], int type)
{
	DeQuantMatrix(M, QTable[type ? 1 : 0]);
	TransposeMatrix(M);
	IDQT(M);
}

void TransposeMatrix(int M[][8])
{
	int T[8][8];
	int *t = (int *)M;
	int *p = (int *)&T;
	int *q = (int *)&TransposeTable;
	for (int i = 0; i < 64; i++) *p++ = t[*q++];
	memcpy(M, T, sizeof(T));
}

void DeQuantMatrix(int M[][8], int Q[][8])
{
	int *m = (int *)M, *q = (int *)Q;
	for (int i = 0; i < 64; i++) *m++ *= *q++;
}

void IDQT(int M[][8])
{
	double T[8][8], t;
	int i, j, k;

	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++)
		{
			t = 0;
			for (k = 0; k < 8; k++)
				t += C[k][i] * M[k][j];
			T[i][j] = t;
		}

	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++)
		{
			t = 0;
			for (k = 0; k < 8; k++)
				t += T[i][k] * C[k][j];
			M[i][j] = (int)t;
		}
}

